@@ -24,6 +24,7 @@
 #endif
 
 #include <string.h>
+#include <gnome.h>
 #include "gok-key.h"
 #include "gok-keyboard.h"
 #include "gok-mousecontrol.h"
@@ -95,6 +96,10 @@ gok_key_lookup_shorter_label (char *str)
 		g_free (str);
 		str = "Left\nTab";
 	}
+	else if (!strcmp (str, "Sterling")) {
+		g_free (str);
+		str = "£";
+	}
 	else {
 		gint len;
 		gboolean prev_is_lower, is_upper;
@@ -331,7 +336,7 @@ gok_key_set_effective_group (gint group)
 }
 
 void
-gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec)
+gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec, gchar *align)
 {
 	if (xmlStrPrefix (sizespec, XML_STRING_SIZE_STOCK)) {
 		keyimage->type = IMAGE_TYPE_STOCK;
@@ -343,13 +348,18 @@ gok_keyimage_set_size_from_spec (GokKeyImage *keyimage, gchar *sizespec)
 			keyimage->w = w;
 			keyimage->h = h;
 		}
-		fprintf (stderr, "image size: %d,%d\n", keyimage->w, keyimage->h);
 	}
 	else if (xmlStrPrefix (sizespec, XML_STRING_SIZE_FIT))
 		keyimage->type = IMAGE_TYPE_FIT;
 	else if (xmlStrPrefix (sizespec, XML_STRING_SIZE_FILL))
 		keyimage->type = IMAGE_TYPE_FILL;
 
+	if (!xmlStrcmp (align, (const xmlChar *) "right")) {
+		keyimage->placement_policy = IMAGE_PLACEMENT_RIGHT;
+	}
+	else { /* TODO: that's all we recognize for now, implement other options */
+		keyimage->placement_policy = IMAGE_PLACEMENT_LEFT;
+	}
 }
 
 GokKey *
@@ -598,6 +608,7 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 		(bDeleteButton == TRUE))
 	{
 		gtk_widget_destroy (pKey->pButton);
+		pKey->pButton = NULL;
 	}
 	
 	/* delete all the outputs associated with the key */
@@ -606,6 +617,9 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 	gok_output_delete_all (pKey->pOutputWrapperPost);
 
 	g_free (pKey);
+
+	gok_feedback_drop_refs (pKey);
+	gok_scanner_drop_refs (pKey);
 }
 
 /**
@@ -684,8 +698,8 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchEditText") == 0)
 	{
-		pKey->Type = KEYTYPE_BRANCHEDITTEXT;
-		pKey->Style = KEYSTYLE_BRANCHEDITTEXT;
+		pKey->Type = KEYTYPE_BRANCHCOMPOSE;
+		pKey->Style = KEYSTYLE_GENERALDYNAMIC;
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchSettings") == 0)
 	{
@@ -703,11 +717,9 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 			gok_gconf_get_bool (gconf_client_get_default(),
 				GOK_GCONF_PREFS_LOCKED, &settings_locked);
 			if (settings_locked) {
-				pKey->Style = KEYSTYLE_INSENSITIVE;
-			}
-			else {
-				pKey->Style = KEYSTYLE_SETTINGS;
+				gok_settingsdialog_lock ();
 			}
+			pKey->Style = KEYSTYLE_SETTINGS;
 		}
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"branchWindows") == 0)
@@ -759,6 +771,35 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_ABOUT;
 		pKey->Style = KEYSTYLE_ABOUT;
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"navigate") == 0)
+	{
+		pKey->Type = KEYTYPE_TEXTNAV;
+		pKey->Style = KEYSTYLE_NORMAL;
+		pKey->is_repeatable = TRUE;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"edit") == 0)
+	{
+		pKey->Type = KEYTYPE_EDIT;
+		pKey->Style = KEYSTYLE_NORMAL;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"select") == 0)
+	{
+		pKey->Type = KEYTYPE_SELECT;
+		pKey->Style = KEYSTYLE_SELECT;
+		gok_compose_key_init (pKey, pNode);
+	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"select-toggle") == 0)
+	{
+		pKey->Type = KEYTYPE_TOGGLESELECT;
+		pKey->Style = KEYSTYLE_SELECT;
+		pKey->pImage = gok_keyimage_new (pKey, NULL);
+		pKey->pImage->type = IMAGE_TYPE_INDICATOR;
+		pKey->has_image = TRUE;
+		pKey->ComponentState.active = FALSE;
+		gok_compose_key_init (pKey, pNode);
+	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"repeat-next") == 0)
 	{
 		pKey->Type = KEYTYPE_REPEATNEXT;
@@ -886,9 +927,10 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 			/* TODO: support href as well as local files ? */
 			gchar *filename = xmlGetProp (pNodeKeyChild, (const xmlChar *) "source");
 			gchar *sizespec = xmlGetProp (pNodeKeyChild, (const xmlChar *) "type");
+			gchar *align = xmlGetProp (pNodeKeyChild, (const xmlChar *) "align");
 			gok_keyimage_new (pKey, filename);
 			if (sizespec != NULL) 
-				gok_keyimage_set_size_from_spec (pKey->pImage, sizespec);
+				gok_keyimage_set_size_from_spec (pKey->pImage, sizespec, align);
 			pKey->has_image = TRUE;
 		}
 
@@ -1241,16 +1283,16 @@ gok_key_status_image (GokKey *key)
 
 #if STATUS_IMAGES
 	if (state.latched) {
-		icon_name = "gok/latched.png";
+		icon_name = "latched.png";
 	}
 	else if (state.locked) {
-		icon_name = "gok/locked.png";
+		icon_name = "locked.png";
 	}
 	else {
-		icon_name = "gok/empty.png";
+		icon_name = "empty.png";
 	}
 #else
-	icon_name = "gok/small-empty.png";
+	icon_name = "small-empty.png";
 #endif
 	file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_APP_DATADIR,
 					  icon_name,
@@ -1284,29 +1326,31 @@ gok_key_create_image_widget (GokKey* pKey)
 			break;
 		case IMAGE_TYPE_STOCK:
 			image = gtk_image_new_from_stock (pKey->pImage->Filename,
-							  pKey->pImage->stock_size);
+							  GTK_ICON_SIZE_BUTTON /*pKey->pImage->stock_size*/);
 			break;
 		case IMAGE_TYPE_FIXED:
 		default:
-			fprintf (stderr, "scaled image %d,%d\n", 
-				 pKey->pImage->w, pKey->pImage->h);
 			if ((pKey->pImage->w == -1) && (pKey->pImage->h == -1)) {
-				image = gtk_image_new_from_file (
-					gok_key_get_image_filename (pKey));
+				file = gok_key_get_image_filename (pKey);
+				image = gtk_image_new_from_file (file);
+				g_free (file);
 			}
 			else {
 				GdkPixbuf *scaled = NULL, *pixbuf;
-				pixbuf = gdk_pixbuf_new_from_file (
-					gok_key_get_image_filename (pKey), NULL);
-				if (pixbuf)
+				file = gok_key_get_image_filename (pKey);
+				pixbuf = gdk_pixbuf_new_from_file (file, NULL);
+				g_free (file);
+				if (pixbuf) {
 					scaled = gdk_pixbuf_scale_simple (
 						pixbuf, pKey->pImage->w,
 						pKey->pImage->h, 
 						GDK_INTERP_BILINEAR);
-				g_object_unref (pixbuf);
-				if (scaled)
+					g_object_unref (pixbuf);
+				}
+				if (scaled) {
 					image = gtk_image_new_from_pixbuf (scaled);
-				g_object_unref (scaled);
+					g_object_unref (scaled);
+				}
 			}
 			break;
 		}
@@ -1508,11 +1552,16 @@ void gok_key_set_button_name (GokKey* pKey)
 			strcpy (styleText, "StyleTextNormal");
 		}
 		break;
-		
+
 	case KEYSTYLE_GENERALDYNAMIC:
 		strcpy (styleButton, "StyleButtonGeneralDynamic");
 		strcpy (styleText, "StyleTextNormal");
 		break;
+
+	case KEYSTYLE_SELECT:
+		strcpy (styleButton, "StyleButtonSelect");
+		strcpy (styleText, "StyleTextNormal");
+		break;
 		
 	case KEYSTYLE_BRANCHMENUS:
 		strcpy (styleButton, "StyleButtonBranchMenus");
@@ -1541,8 +1590,9 @@ void gok_key_set_button_name (GokKey* pKey)
 		break;
 
 	case KEYSTYLE_BRANCHGUIACTIONS:
+	case KEYSTYLE_PAGESELECTION:
 	case KEYSTYLE_HYPERLINK:
-	case KEYSTYLE_EDITABLETEXTACTION:
+	case KEYSTYLE_EDIT:
 		strcpy (styleButton, "StyleButtonBranchGuiActions");
 		strcpy (styleText, "StyleTextNormal");
 		break;
@@ -1565,6 +1615,7 @@ void gok_key_set_button_name (GokKey* pKey)
 					
 	case KEYSTYLE_POINTERCONTROL:
 	case KEYSTYLE_WORDCOMPLETE:
+	case KEYSTYLE_SPELL:
 		strcpy (styleButton, "StyleButtonWordComplete");
 		strcpy (styleText, "StyleTextWordComplete");
 		break;
@@ -2065,6 +2116,7 @@ gok_keyimage_new (GokKey* pKey, gchar* pFilename)
 	GokKeyImage* pNewImage;
 	GokKeyImage* pKeyImage;
 
+	gok_log_enter();
 	pNewImage = (GokKeyImage*) g_malloc (sizeof (GokKeyImage));
 
 	pNewImage->Filename = NULL;
@@ -2096,6 +2148,7 @@ gok_keyimage_new (GokKey* pKey, gchar* pFilename)
 		pKey->has_image = TRUE;
 	}
 	
+	gok_log_leave();
 	return pNewImage;
 }
 	 
@@ -2119,10 +2172,13 @@ gok_keyimage_delete (GokKeyImage* pKeyImage)
 gchar*
 gok_key_get_image_filename (GokKey* pKey)
 {
+	gchar *filename = NULL;
 	g_assert (pKey != NULL);
-	
-	/* this needs to get the filename for the normal modifier key */
-	return pKey->pImage->Filename;
+
+	filename = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_APP_DATADIR,
+					      pKey->pImage->Filename,
+					      FALSE, NULL);
+	return filename;
 }
 
 /**
